# -*- coding: UTF-8 -*-
"""Dummy Codecs - simple string manipulations.

These are dummy codecs for manipulating strings, for use with other codecs in encoding/decoding chains.

These codecs:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)
"""
import re

from ..__common__ import add


def replace(pair, *args):
    def code(input, error="strict"):
        return input.replace(pair[0], pair[1]), len(input)
    return code
add("replace", replace, replace, r"^replace[-_]?((?!.*(.).*\2)..)$", guess=None)
# important note:                                              ^
#                                           using "{2}" here instead will break the codec
#  this is due to the fact the codext.__common__.generate_string_from_regex DOES NOT handle ASSERT_NOT (?!) and will
#   faill to generate a valid instance in lookup(...) when an encoding name is to be generated to get the CodecInfo


def substitute(token, replacement):
    def code(input, error="strict"):
        return input.replace(token, replacement), len(input)
    return code
add("substitute", substitute, substitute, r"^substitute[-_]?(.*?)/(.*?)$", guess=None)


reverse = lambda i, e="strict": (i[::-1], len(i))
add("reverse", reverse, reverse)

word_reverse = lambda i, e="strict": (" ".join(w[::-1] for w in i.split()), len(i))
add("reverse-words", word_reverse, word_reverse, r"^reverse[-_]words$")

strip_spaces = lambda i, e="strict": (i.replace(" ", ""), len(i))
add("strip-spaces", strip_spaces, strip_spaces, guess=None)




# -*- coding: UTF-8 -*-
"""Pkzip Codec - pkzip content compression.

NB: Not an encoding properly speaking.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)
"""
import zipfile

from ..__common__ import *


_str          = ["test", "This is a test", "@random{512,1024,2048}"]
__examples1__ = {'enc-dec(pkzip-deflate|deflate)': _str}
__examples2__ = {'enc-dec(pkzip_bz2|bzip2)':       _str}
__examples3__ = {'enc-dec(pkzip-lzma|lzma)':       _str}


if PY3:
    NULL = {
        8:  b"\x03\x00",
        12: b"BZh9\x17rE8P\x90\x00\x00\x00\x00",
        14: b"\t\x04\x05\x00]\x00\x00\x80\x00\x00\x83\xff\xfb\xff\xff\xc0\x00\x00\x00",
    }    
        
    
    def pkzip_encode(compression_type):
        def _encode(text, errors="strict"):
            c = zipfile._get_compressor(compression_type)
            return c.compress(b(text)) + c.flush(), len(text)
        return _encode


    def pkzip_decode(compression_type, name):
        def _decode(data, errors="strict"):
            d = zipfile._get_decompressor(compression_type)
            r = d.decompress(b(data))
            if len(r) == 0 and b(data) != NULL[compression_type]:
                return handle_error(name, errors, decode=True)(data[0], 0) if len(data) > 0 else "", len(data)
            return r, len(r)
        return _decode


    add("pkzip_deflate", pkzip_encode(8), pkzip_decode(8, "deflate"), r"(?:(?:pk)?zip[-_])?deflate",
        entropy=7.9, examples=__examples1__, guess=["deflate"])

    add("pkzip_bzip2", pkzip_encode(12), pkzip_decode(12, "bzip2"), r"(?:(?:pk)?zip[-_])?bz(?:ip)?2",
        entropy=7.9, examples=__examples2__, guess=["bz2"])

    add("pkzip_lzma", pkzip_encode(14), pkzip_decode(14, "lzma"), r"(?:(?:pk)?zip[-_])?lzma",
        entropy=7.9, examples=__examples3__, guess=["lzma"])




# -*- coding: UTF-8 -*-
"""Case Codecs - string hashing with Unix's Crypt.

These are codecs for hashing strings, for use with other codecs in encoding chains.

These codecs:
- transform strings from str to str
- transform strings from bytes to bytes
- transform file content from str to bytes (write)
"""
from ..__common__ import add, ensure_str, PY3, UNIX


if PY3 and UNIX:
    import crypt
    
    METHODS = [x[7:].lower() for x in crypt.__dict__ if x.startswith("METHOD_")]
    
    def crypt_hash(method):
        method = (method or "").lstrip("-_") or "blowfish"
        if method not in METHODS:
            raise NotImplementedError("method '%s' is not implemented" % method)
        def _encode(input, error="strict"):
            m = getattr(crypt, "METHOD_" + method.upper())
            return crypt.crypt(ensure_str(input), crypt.mksalt(m)), len(input)
        return _encode
    
    add("crypt", crypt_hash, pattern=r"^crypt(|[-_](?:%s))$" % "|".join(METHODS), guess=None)




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""Codecs added assets' tests.

"""
import codecs
import codext
import json
import random
import sys
from codext.__common__ import PERS_MACROS, PERS_MACROS_FILE
from six import b, binary_type, text_type
from unittest import TestCase


PY3 = sys.version[0] == "3"


def dummy_encode(input, errors="strict"):
    return input, len(input)


def dummy_decode(input, errors="strict"):
    return input, len(input)


def dummy_errored_decode(useless):
    raise AttributeError
    def decode(input, errors="strict"):
        return input, len(input)
    return decode


def ensure_str(s, encoding='utf-8', errors='strict'):
    """ Similar to six.ensure_str. Adapted here to avoid messing up with six version errors. """
    if not PY3 and isinstance(s, text_type):
        return s.encode(encoding, errors)
    elif PY3 and isinstance(s, binary_type):
        try:
            return s.decode(encoding, errors)
        except:
            return s.decode("latin-1")
    return s


def getregentry(encoding):
    if encoding == "dummy3":
        return codecs.CodecInfo(name="dummy3", encode=dummy_encode, decode=dummy_decode)


class TestCommon(TestCase):
    def setUp(self):
        codext.reset()
    
    def test_add_codec(self):
        self.assertRaises(ValueError, codext.add, "test")
        self.assertRaises(ValueError, codext.add, "test", "BAD")
        self.assertRaises(ValueError, codext.add, "test", lambda: None, "BAD")
        self.assertIsNone(codext.add("dummy", dummy_encode, dummy_decode))
        self.assertEqual(codext.encode("test", "dummy"), "test")
        ci = codext.lookup("dummy")
        for k in ["add_to_codecs", "category", "examples", "name", "pattern", "text"]:
            self.assertIn(k, ci.parameters.keys())
        self.assertIsNone(codext.add("dummy_errored", None, dummy_errored_decode, r"dummy_errored(\d+)$"))
        self.assertRaises(AttributeError, codext.lookup, "dummy_errored1")

    def test_add_map_codec(self):
        ENCMAP = [{'a': "A", 'b': "B", 'c': "C"}, {'d': "D", 'e': "E", 'f': "F"}, {'g': "G", 'h': "H", 'i': "I"}]
        self.assertIsNone(codext.add_map("dummy2", ENCMAP, pattern=r"^dummy2(?:[-_]?(\d))?$"))
        self.assertRaises(ValueError, codext.add_map, "dummy2", "BAD_ENCMAP")
        self.assertEqual(codext.encode("abc", "dummy2"), "ABC")
        self.assertEqual(codext.encode("abc", "dummy2-1"), "ABC")
        self.assertEqual(codext.encode("def", "dummy2-2"), "DEF")
        self.assertEqual(codext.encode("ghi", "dummy2-3"), "GHI")
        self.assertRaises(LookupError, codext.encode, "test", "dummy2-4")
        ENCMAP = {'': {'a': "A", 'b': "B"}, r'bad': {'a': "B", 'b': "A"}}
        self.assertIsNone(codext.add_map("dummy3", ENCMAP, pattern=r"^dummy3([-_]inverted)?$"))
        self.assertRaises(LookupError, codext.encode, "test", "dummy3_inverted")
        self.assertRaises(ValueError, codext.add_map, "dummy2", ENCMAP, ignore_case="BAD")
        self.assertRaises(ValueError, codext.add_map, "dummy2", ENCMAP, intype="BAD")
        self.assertRaises(ValueError, codext.add_map, "dummy2", ENCMAP, outype="BAD")
        ci = codext.lookup("dummy2")
        for k in ["category", "encmap", "ignore_case", "intype", "no_error", "outype", "repl_char", "sep", "text"]:
            self.assertIn(k, ci.parameters.keys())
    
    def test_list_codecs(self):
        self.assertTrue(len(codext.list()) > 0)
        self.assertTrue(len(codext.list("other")) > 0)
        self.assertTrue(len(codext.list("native")) > 0)
        self.assertTrue(len(codext.list("non-native")) > 0)
        self.assertTrue(len(codext.list("native", "non-native", "crypto", "base")) > 0)
        self.assertTrue(len(codext.list("native", "language", "crypto")) > 0)
        self.assertEqual(set(codext.list()), set(codext.list("native") + codext.list("non-native")))
        self.assertRaises(ValueError, codext.list, "BAD_CATEGORY")
        self.assertTrue(codext.is_native("base64_codec"))
        self.assertFalse(codext.is_native("base64"))
    
    def test_remove_codec(self):
        self.assertIsNone(codext.add("dummy", dummy_encode, dummy_decode))
        self.assertEqual(codext.encode("test", "dummy"), "test")
        self.assertIsNone(codext.remove("dummy"))
        self.assertRaises(LookupError, codext.encode, "test", "dummy")
        # special case, when adding a new codec also to the native codecs registry, then it won't be possible to remove
        #  it afterwards
        self.assertIsNone(codecs.add("dummy2", dummy_encode, dummy_decode))
        self.assertEqual(codecs.encode("test", "dummy2"), "test")
        self.assertIsNone(codecs.remove("dummy2"))
        self.assertEqual(codecs.encode("test", "dummy2"), "test")
        self.assertIsNone(codecs.register(getregentry))
        self.assertEqual(codecs.encode("test", "dummy3"), "test")
        self.assertIsNone(codecs.remove("dummy3"))
        self.assertEqual(codecs.encode("test", "dummy3"), "test")
    
    def test_clear_codecs(self):
        self.assertIsNotNone(codecs.encode("test", "morse"))
        self.assertIsNone(codecs.clear())
        self.assertRaises(LookupError, codecs.encode, "test", "morse")
    
    def test_reset_codecs(self):
        self.assertIsNone(codext.reset())
        self.assertIsNotNone(codext.encode("test", "morse"))
        self.assertRaises(LookupError, codext.encode, "test", "dummy")
    
    def test_search_codecs(self):
        self.assertIsNotNone(codext.search("morse"))
        self.assertIsNotNone(codext.search("geohash"))
        self.assertIsNotNone(codext.examples("morse"))
        self.assertIsNotNone(codext.examples("cp"))
        self.assertIsNotNone(list(codext.generate_strings_from_regex(r"[ab]{1,3}")))
        self.assertIsNotNone(list(codext.generate_strings_from_regex(r"(?<=ab)cd")))
        self.assertIsNotNone(list(codext.generate_strings_from_regex(r"(?<=-)\w+")))
        self.assertIsNotNone(list(codext.generate_strings_from_regex(r"([^\s])\1")))
        self.assertIsNotNone(list(codext.generate_strings_from_regex(r"[^\\]")))
        self.assertIsNotNone(list(codext.generate_strings_from_regex(r"[^a]")))
    
    def test_encode_multiple_rounds(self):
        if PY3:
            self.assertRaises(TypeError, codext.encode, b"test", "utf-8[2]")
        s = "test"
        for i in range(3):
            s = codext.encode(s, "morse")
        self.assertEqual(s, codext.encode("test", "morse[3]"))
        self.assertIsNotNone(codext.encode("test", "base64[10]"))
    
    def test_guess_decode(self):
        self.assertIsNone(codext.stopfunc._reload_lang())
        _l = lambda d: list(d.items())[0][1] if len(d) > 0 else None
        codext.add("test_codec", lambda x, e="strict": (x + "=", len(x)), lambda x, e="strict": (x[:-1], len(x)-1),
                   "test", no_error=True, bonus_func=lambda *a: True, penalty=-.5)
        self.assertIn("test-codec", codext.list_encodings("test"))
        self.assertEqual(codext.decode("TEST=", "test"), "TEST")
        self.assertEqual(list(codext.guess("TEST=", codext.stopfunc.text, codec_categories="test", max_depth=2,
                                           scoring_heuristic=False).items())[0][1], "TEST")
        self.assertEqual(list(codext.guess("TEST=", codext.stopfunc.text, codec_categories=["test", "base"],
                                           max_depth=2).items())[0][1], "TEST")
        STR = "This is a test"
        self.assertEqual(STR, _l(codext.guess("VGhpcyBpcyBhIHRlc3Q=", "a test", max_depth=1)))
        self.assertEqual(STR, _l(codext.guess("CJG3Ix8bVcSRMLOqwDUg28aDsT7", "a test", found=["base62"])))
        self.assertEqual(STR, _l(codext.guess("VGhpcyBpcyBhIHRlc3Q=", "a test", 0, 1, "base", scoring_heuristic=True,
                                              exclude=["base100"])))
        self.assertEqual(STR, _l(codext.guess("VGhpcyBpcyBhIHRlc3Q=", "a test", 0, 1, ["base", "crypto"])))
        self.assertEqual(len(codext.guess("NOT THE ENCODED TEST STRING", "a test", max_depth=1, exclude=[None])), 0)
        self.assertIn("F1@9", _l(codext.guess("VGVzdCBGMUA5ICE=", codext.stopfunc.flag, max_depth=1, stop=False,
                                              show=True)))
        self.assertEqual(len(codext.guess("VGhpcyBpcyBhIHRlc3Q=", " a test", max_depth=1, codec_categories="base",
                                          exclude=("base64", "base64-url"))), 0)
        self.assertEqual(len(codext.guess("VGhpcyBpcyBhIHRlc3Q=", " a test", max_depth=1, codec_categories="base",
                                          scoring_heuristic=True, exclude=("base64", "base64-url", "atbash"))), 0)
        self.assertRaises(ValueError, codext.guess, STR, max_depth=0)
        self.assertRaises(ValueError, codext.guess, STR, exclude=42)
        for c in ["base", "language", "native", "stegano"]:
            e = codext.list(c)
            random.shuffle(e)
            for ename in e[:10]:
                for encoding in codext.lookup(ename).parameters.get('guess', [ename])[:10]:
                    try:
                        enc = codext.encode(STR, encoding)
                    except (NotImplementedError, ValueError):
                        continue
                    except TypeError:
                        enc = codext.encode(b(STR), encoding)
                    if codext.decode(enc, encoding) == STR:
                        continue
                    for found_encodings, found_dec in codext.guess(enc, "a test", 0, 1, [c],
                                                                   scoring_heuristic=True, debug=True).items():
                        self.assertEqual(ensure_str(STR).lower(), ensure_str(found_dec).lower())
                        if c != "base":
                            # do not check for base as the guessed encoding name can be different, e.g.:
                            #  actual:  base2
                            #  guessed: base2-generic
                            if "-icase" in encoding:
                                self.assertEqual(encoding.lower(), found_encodings[0].lower())
                            else:
                                self.assertEqual(encoding, found_encodings[0])
        txt = "".join(chr(i) for i in range(256))
        b64 = codext.encode(txt, "base64")
        self.assertEqual(txt, _l(codext.guess(b64, "0123456789", max_depth=1, scoring_heuristic=True,
                                              codec_categories="base")))
        self.assertRaises(ValueError, codext.stopfunc._reload_lang, "DOES_NOT_EXIST")
    
    def test_rank_input(self):
        codext.add("test_codec", lambda x, e="strict": (x + "=", len(x)), lambda x, e="strict": (x[:-1], len(x)-1),
                   "test", no_error=True, bonus_func=lambda *a: True, penalty=-.5)
        STR = "This is a test string !"
        ENC = codext.encode(STR, "base64")
        self.assertTrue(len(codext.rank(ENC)) > 20)
        self.assertEqual(len(codext.rank(ENC, limit=20)), 20)
        self.assertEqual(codext.rank(ENC, exclude=["rot"])[0][1], "base64")
        self.assertEqual(codext.rank(ENC, codec_categories="base")[0][0][1], STR)
        self.assertEqual(codext.rank(ENC, codec_categories=["base"])[0][0][1], STR)
        self.assertIsNotNone(codext.rank(ENC, codec_categories=["base"], exclude=["does_not_exist"])[0][0][1], STR)
        self.assertIsNotNone(codext.rank("TEST=", codec_categories=["test", "base"])[0][0][1], "TEST")
    
    def test_handle_macros(self):
        MACRO = "test-macro-f2ca1bb6c7e907d06dafe4687e579fce76b37e4e93b7605022da52e6ccc26fd2"
        STR = "this is a test"
        ENC = "H4sIAMrbkmEC/0txzyhIrnQC4QxPj6CcZONAWwAMIDOIFAAAAA=="
        codext.remove(MACRO)
        l = codext.list_macros()
        self.assertTrue(len(l) > 0)
        cm = codext.lookup("example-macro")
        self.assertIsNotNone(cm)
        self.assertRaises(LookupError, codext.lookup, "example-macro", False)
        self.assertRaises(ValueError, codext.add_macro, "example-macro", "base64")
        self.assertRaises(ValueError, codext.add_macro, "base64", "base91")
        self.assertIsNotNone(repr(cm))
        self.assertTrue(hasattr(cm, "parameters"))
        self.assertRaises(LookupError, codext.lookup, MACRO)
        self.assertIsNone(codext.add_macro(MACRO, "base64", "gzip", "base64"))
        self.assertIn(MACRO, codext.list_macros())
        self.assertIsNotNone(codext.encode(STR, MACRO))
        self.assertEqual(codext.decode(ENC, MACRO), STR)
        # insert a bad entry for the list of encodings in the JSON file
        PERS_MACROS[MACRO] = "not a list or tuple..."
        with open(PERS_MACROS_FILE, 'w') as f:
            json.dump(PERS_MACROS, f)
        codext.reset()
        self.assertRaises(ValueError, codext.lookup, MACRO)
        self.assertIsNone(codext.remove(MACRO))
        self.assertRaises(LookupError, codext.lookup, MACRO)
        self.assertNotIn(MACRO, codext.list_macros())
        self.assertIsNone(codext.remove("THIS-MACRO-DOES-NOT-EXIST"))
        self.assertIsNone(codext.remove("VALID-MACRO"))
        self.assertIsNone(codext.add_macro("VALID-MACRO", "gzip", "base64"))
        self.assertIsNone(codext.remove("VALID-MACRO"))
        if PY3:
            self.assertIsNone(codext.add_macro("VALID-MACRO", "lzma", "base64"))
            self.assertIsNone(codext.remove("VALID-MACRO"))
        self.assertRaises(ValueError, codext.add_macro, "SHALL-FAIL", "base26", "sms", "letter-indices")




# -*- coding: UTF-8 -*-
"""Case Codecs - simple string case manipulations.

These are case-related codecs for manipulating strings, for use with other codecs in encoding/decoding chains.

These codecs:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)
"""
import re

from ..__common__ import add


pascal = lambda i, e="strict": ("".join(x.capitalize() for x in re.findall(r"[0-9a-z]+", i.lower())), len(i))
add("camel", lambda i, e: uncapitalize(pascal(i, e)[0]), None, r"^camel(?:[-_]?case)?$")
add("pascal", pascal, None, r"^pascal(?:[-_]?case)?$")

capitalize = lambda i, e="strict": (i.capitalize(), len(i))
uncapitalize = lambda i, e="strict": (i[0].lower() + i[1:] if len(i) > 0 else "", len(i))
add("capitalize", capitalize, uncapitalize)

lowercase = lambda i, e="strict": (i.lower(), len(i))
uppercase = lambda i, e="strict": (i.upper(), len(i))
add("uppercase", uppercase, lowercase, r"^upper(?:case)?$")
add("lowercase", lowercase, uppercase, r"^lower(?:case)?$")

slugify = lambda i, e="strict", d="-": (re.sub(r"[^0-9a-z]+", d, i.lower()).strip(d), len(i))
add("slugify", lambda i, e: slugify(i, e), None, r"^(?:slug(?:ify)?|kebab(?:[-_]?case)?)$")
add("snake", lambda i, e: slugify(i, e, "_"), None, r"^snake(?:[-_]?case)$")

swapcase = lambda i, e="strict": (i.swapcase(), len(i))
add("swapcase", swapcase, swapcase, r"^swap(?:[-_]?case)?$")

title = lambda i, e="strict": (i.title(), len(i))
untitle = lambda i, e="strict": (" ".join(w[0].lower() + w[1:] if len(w) > 0 else "" for w in i.split()), len(i))
add("title", title, untitle)




# -*- coding: UTF-8 -*-
"""Case Codecs - string hashing with Secure Hash Algorithms.

These are codecs for hashing strings, for use with other codecs in encoding chains.

These codecs:
- transform strings from str to str
- transform strings from bytes to bytes
- transform file content from str to bytes (write)
"""
import hashlib

from ..__common__ import add, b, PY3


add("sha1", lambda s, error="strict": (hashlib.sha1(b(s)).hexdigest(), len(s)), guess=None)
add("sha224", lambda s, error="strict": (hashlib.sha224(b(s)).hexdigest(), len(s)), guess=None)
add("sha256", lambda s, error="strict": (hashlib.sha256(b(s)).hexdigest(), len(s)), guess=None)
add("sha384", lambda s, error="strict": (hashlib.sha384(b(s)).hexdigest(), len(s)), guess=None)
add("sha512", lambda s, error="strict": (hashlib.sha512(b(s)).hexdigest(), len(s)), guess=None)


if PY3:
    add("sha3_224", lambda s, error="strict": (hashlib.sha3_224(b(s)).hexdigest(), len(s)), pattern=r"^sha3[-_]224$",
        guess=None)
    add("sha3_256", lambda s, error="strict": (hashlib.sha3_256(b(s)).hexdigest(), len(s)), pattern=r"^sha3[-_]256$",
        guess=None)
    add("sha3_384", lambda s, error="strict": (hashlib.sha3_384(b(s)).hexdigest(), len(s)), pattern=r"^sha3[-_]384$",
        guess=None)
    add("sha3_512", lambda s, error="strict": (hashlib.sha3_512(b(s)).hexdigest(), len(s)), pattern=r"^sha3[-_]512$",
        guess=None)




# -*- coding: UTF-8 -*-
"""Case Codecs - string hashing with blake.

These are codecs for hashing strings, for use with other codecs in encoding chains.

These codecs:
- transform strings from str to str
- transform strings from bytes to bytes
- transform file content from str to bytes (write)
"""
import hashlib

from ..__common__ import add, b, PY3


if PY3:
    def blake_hash(c):
        def _hash_transform(l):
            l = (l or "64" if c == "b" else "32").lstrip("_-")
            def _encode(data, error="strict"):
                return getattr(hashlib, "blake2%s" % c)(b(data), digest_size=int(l)).hexdigest(), len(data)
            return _encode
        return _hash_transform

    add("blake2b", blake_hash("b"), pattern=r"^blake2b(|[-_](?:[1-9]|[1-5]\d|6[0-4]))$", guess=None)
    add("blake2s", blake_hash("s"), pattern=r"^blake2s(|[-_](?:[1-9]|[1-2]\d|3[0-2]))$", guess=None)




# -*- coding: UTF-8 -*-
"""Markdown Codec - markdown content conversion to HTML.

This codec:
- encodes strings from str to str
- encodes strings from bytes to bytes
- encodes file content from str to bytes (write)
"""
from ..__common__ import *


__guess__ = []


try:
    from markdown2 import markdown as md2html
    # note: the group is NOT captured so that the pattern is only used to match the name of the codec and not to
    #        dynamically bind to a parametrizable encode function
    add("markdown", lambda md, error="strict": (md2html(md), len(md)), pattern=r"^(?:markdown|Markdown|md)$")
except ImportError:
    pass




# -*- coding: UTF-8 -*-
"""ROT Codec - rot-with-N-offset content encoding.

This codec:
- en/decodes strings from str to str
- en/decodes strings from bytes to bytes
- decodes file content to str (read)
- encodes file content from str to bytes (write)
"""
from string import ascii_lowercase as LC, ascii_uppercase as UC, digits as DIG

from ..__common__ import *


__examples1__ = {
    'enc(rot0|rot--10|rot100)': None,
    'enc(rot1|rot-1|caesar_1)': {'this is a test': "uijt jt b uftu"},
    'enc(rot3|caesar-3)':       {'this is a test': "wklv lv d whvw"},
    'enc(rot47)':               {'this is a test': "E9:D :D 2 E6DE"},
}
__examples2__ = {
    'enc(prot0|prot--10|prot100)': None,
    'enc(prot1|prog-caesar_1)':    {'this is a test': "ujlw oz j eqfh"},
    'enc(prot3|pcaesar-3)':        {'this is a test': "wlny qb l gshj"},
}
__examples3__ = {
    'enc(arot0|arot--10|arot100)': None,
    'enc(arot1|alt-caesar_1)':     {'this is a test': "ugjr ht b udts"},
    'enc(arot3|acaesar-3)':        {'this is a test': "welp fv d wbvq"},
}
__guess1__ = ["rot-%d" % i for i in range(1, 26)] + ["rot-47"]
__guess2__ = ["progressive-rot-%d" % i for i in range(1, 26)] + ["progressive-rot-n%d" % i for i in range(1, 26)]
__guess3__ = ["alternative-rot-%d" % i for i in range(1, 26)]


ROT47 = "!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~"


def _rotn(text, n=13, a=(LC, UC), alt=False, prog=False, neg=False):
    r = ""
    for i, c in enumerate(ensure_str(text)):
        found = False
        for l in a:
            if c in l:
                r += l[(l.index(c) + [1, -1][alt and i % 2 == 1] * n + ([1, -1][neg] * i if prog else 0)) % len(l)]
                found = True
                break
        if not found:
            r += c
    return r


def arot_encode(i):
    def encode(text, errors="strict"):
        return _rotn(ensure_str(text), i, alt=True), len(text)
    return encode


def arot_decode(i):
    def decode(text, errors="strict"):
        return _rotn(ensure_str(text), -i, alt=True), len(text)
    return decode


def rot_encode(i):
    def encode(text, errors="strict"):
        t = ensure_str(text)
        r = _rotn(t, 47, [ROT47]) if i == 47 else _rotn(t, i)
        return r, len(r)
    return encode


def rot_decode(i):
    def decode(text, errors="strict"):
        t = ensure_str(text)
        r = _rotn(t, -47, [ROT47]) if i == 47 else _rotn(t, -i)
        return r, len(r)
    return decode


def prot_encode(n, i):
    def encode(text, errors="strict"):
        return _rotn(ensure_str(text), i, prog=True, neg=n == "n"), len(text)
    return encode


def prot_decode(n, i):
    def decode(text, errors="strict"):
        return _rotn(ensure_str(text), -i, prog=True, neg=n != "n"), len(text)
    return decode


add("alternative-rot", arot_encode, arot_decode, r"a(?:lt(?:ernative)?-)?(?:caesar|rot)[-_]?([1-9]|1[0-9]|2[0-5])$",
    penalty=.2, entropy=lambda e: e, printables_rate=lambda pr: pr, transitive=True, examples=__examples3__,
    guess=__guess3__)
add("rot", rot_encode, rot_decode, r"(?:caesar|rot)[-_]?([1-9]|1[0-9]|2[0-5]|47)$", aliases=["caesar"], penalty=.2,
    entropy=lambda e: e, printables_rate=lambda pr: pr, transitive=True, examples=__examples1__, guess=__guess1__)
add("progressive-rot", prot_encode, prot_decode, r"p(?:rog(?:ressive)?-)?(?:caesar|rot)[-_]?(n?)([1-9]|1[0-9]|2[0-5])$",
    penalty=.2, entropy=lambda e: e, printables_rate=lambda pr: pr, transitive=True, examples=__examples2__,
    guess=__guess2__)




# -*- coding: UTF-8 -*-
"""Codecs extension module.

"""
from __future__ import print_function
from ast import literal_eval
from six import binary_type, text_type

from .__common__ import *
from .__info__ import __author__, __copyright__, __email__, __license__, __source__, __version__


__all__ = ["add", "add_map", "clear", "decode", "encode", "guess", "lookup",  "open", "rank", "register", "remove",
           "reset"]

decode   = codecs.decode
encode   = codecs.encode
guess    = codecs.guess
lookup   = codecs.lookup
open     = codecs.open

_lst = list
list = list_encodings  # not included in __all__ because of shadow name


reset()


def __literal_eval(o):
    """ Non-failing ast.literal_eval alias function. """
    try:
        return literal_eval(str(o))
    except ValueError:
        return literal_eval("'" + str(o) + "'")


def __print_tabular(lst, space=4):
    try:
        cols, _ = os.get_terminal_size()
        # first, convert the list to a table that fits into the terminal
        i, line, w = 0, "", []
        while i < len(lst):
            x = lst[i]
            l = len(x)
            col = "%-{}s".format(l + space) % x
            i += 1
            w.append(l)
            if len(line) + len(col) > cols:
                break
            line += col
        while True:
            t = [lst[j:j+i] for j in range(0, len(lst), i)]
            w = [max(0 if j+k*i >= len(lst) else len(lst[j+k*i]) for k in range(len(t))) for j, _ in enumerate(w)]
            if sum(w) + space * len(w) >= cols:
                i -= 1
                w.pop()
            else:
                break
        print("\n".join("".join("%-{}s".format(w[n] + space) % x for n, x in enumerate(r)) for r in t) + "\n")
    except (AttributeError, OSError):
        print(", ".join(lst) + "\n")


def main():
    import argparse, os
    descr = "Codecs Extension (CodExt) {}\n\nAuthor   : {} ({})\nCopyright: {}\nLicense  : {}\nSource   : {}\n" \
            "\nThis tool allows to encode/decode input strings/files with an extended set of codecs.\n\n" \
            .format(__version__, __author__, __email__, __copyright__, __license__, __source__)
    examples = "usage examples:\n- " + "\n- ".join([
        "codext search bitcoin",
        "codext decode base32 -i file.b32",
        "codext encode morse < to_be_encoded.txt",
        "echo \"test\" | codext encode base100",
        "echo -en \"test\" | codext encode braille -o test.braille",
        "codext encode base64 < to_be_encoded.txt > text.b64",
        "echo -en \"test\" | codext encode base64 | codext encode base32",
        "echo -en \"mrdvm6teie6t2cq=\" | codext encode upper | codext decode base32 | codext decode base64",
        "echo -en \"test\" | codext encode upper reverse base32 | codext decode base32 reverse lower",
        "echo -en \"test\" | codext encode upper reverse base32 base64 morse",
        "echo -en \"test\" | codext encode base64 gzip | codext guess",
        "echo -en \"test\" | codext encode base64 gzip | codext guess gzip -c base",
    ])
    parser = argparse.ArgumentParser(description=descr, epilog=examples, formatter_class=argparse.RawTextHelpFormatter)
    sparsers = parser.add_subparsers(dest="command", help="command to be executed")
    parser.add_argument("-i", "--input-file", dest="infile", help="input file (if none, take stdin as input)")
    parser.add_argument("-o", "--output-file", dest="outfile", help="output file (if none, display result to stdout)")
    parser.add_argument("-s", "--strip-newlines", action="store_true", dest="strip",
                        help="strip newlines from input (default: False)")
    encode = sparsers.add_parser("encode", help="encode input using the specified codecs")
    encode.add_argument("encoding", nargs="+", help="list of encodings to apply")
    encode.add_argument("-e", "--errors", default="strict", choices=["ignore", "leave", "replace", "strict"],
                        help="error handling (default: strict)")
    decode = sparsers.add_parser("decode", help="decode input using the specified codecs")
    decode.add_argument("encoding", nargs="+", help="list of encodings to apply")
    decode.add_argument("-e", "--errors", default="strict", choices=["ignore", "leave", "replace", "strict"],
                        help="error handling (default: strict)")
    guess = sparsers.add_parser("guess", help="try guessing the decoding codecs")
    guess.add_argument("encoding", nargs="*", help="list of known encodings to apply (default: none)")
    guess.add_argument("-c", "--codec-categories", nargs="*", help="codec categories to be included in the search ; "
                                                                   "format: string|tuple")
    guess.add_argument("-d", "--min-depth", default=0, type=int, help="minimum codec search depth before triggering "
                                                                "results (default: 0)")
    guess.add_argument("-D", "--max-depth", default=5, type=int, help="maximum codec search depth (default: 5)")
    guess.add_argument("-e", "--exclude-codecs", nargs="*", help="codecs to be explicitely not used ; "
                                                                 "format: string|tuple")
    guess.add_argument("-E", "--extended", action="store_true",
                       help="while using the scoring heuristic, also consider null scores (default: False)")
    lng = "lang_%s" % LANG
    def_func = lng if getattr(stopfunc, lng, None) else "text"
    guess.add_argument("-f", "--stop-function", default=def_func, help="result checking function (default: %s) ; "
                       "format: printables|text|flag|lang_[bigram]|[regex]\nNB: [regex] is case-sensitive ; add -i to "
                       "force it as case-insensitive or add '(?i)' in front of the expression" % def_func)
    guess.add_argument("-i", "--case-insensitive", dest="icase", action="store_true",
                       help="while using the regex stop function, set it as case-insensitive (default: False)")
    guess.add_argument("-H", "--no-heuristic", action="store_true", help="DO NOT use the scoring heuristic ; slows down"
                       " the search but may be more accurate (default: False)")
    if len(stopfunc.LANG_BACKENDS) == 0:
        _lb = stopfunc.LANG_BACKEND
        guess.add_argument("-l", "--lang-backend", default=_lb, choices=stopfunc.LANG_BACKENDS + ["none"],
                           help="natural language detection backend (default: %s)" % _lb)
    guess.add_argument("-s", "--do-not-stop", action="store_true",
                       help="do not stop if a valid output is found (default: False)")
    guess.add_argument("-v", "--verbose", action="store_true",
                       help="show guessing information and steps (default: False)")
    rank = sparsers.add_parser("rank", help="rank the most probable encodings based on the given input")
    rank.add_argument("-c", "--codec-categories", help="codec categories to be included in the search ; "
                                                       "format: string|tuple|list(strings|tuples)")
    rank.add_argument("-e", "--exclude-codecs", help="codecs to be explicitely not used ; "
                                                     "format: string|tuple|list(strings|tuples)")
    rank.add_argument("-E", "--extended", action="store_true",
                      help="while using the scoring heuristic, also consider null scores (default: False)")
    rank.add_argument("-l", "--limit", type=int, default=10, help="limit the number of displayed results")
    search = sparsers.add_parser("search", help="search for codecs")
    search.add_argument("pattern", nargs="+", help="encoding pattern to search")
    listi = sparsers.add_parser("list", help="list items")
    lsparsers = listi.add_subparsers(dest="type", help="type of item to be listed")
    liste = lsparsers.add_parser("encodings", help="list encodings")
    liste.add_argument("category", nargs="*", help="selected categories")
    listm = lsparsers.add_parser("macros", help="list macros")
    addm = sparsers.add_parser("add-macro", help="add a macro to the registry")
    addm.add_argument("name", help="macro's name")
    addm.add_argument("encoding", nargs="+", help="list of encodings to chain")
    remm = sparsers.add_parser("remove-macro", help="remove a macro from the registry")
    remm.add_argument("name", help="macro's name")
    args = parser.parse_args()
    try:
        args.codec_categories = _lst(map(__literal_eval, args.codec_categories))
    except (AttributeError, TypeError):
        pass
    try:
        args.exclude_codecs = _lst(map(__literal_eval, args.exclude_codecs))
    except (AttributeError, TypeError):
        pass
    #print(args.codec_categories, args.exclude_codecs)
    try:
        # if a search pattern is given, only handle it
        if args.command == "search":
            results = []
            for enc in args.pattern:
                results.extend(codecs.search(enc))
            print(", ".join(results) or "No encoding found")
            return 0
        # add/remove macros (not requiring to input a file or text)
        elif args.command == "add-macro":
            add_macro(args.name, *args.encoding)
            return 0
        elif args.command == "remove-macro":
            remove_macro(args.name)
            return 0
        # list encodings or macros
        elif args.command == "list":
            if args.type == "encodings":
                cats = args.category or list_categories()
                for c in sorted(cats):
                    l = list_encodings(c)
                    if len(l) > 0:
                        if len(cats) > 0:
                            print(c.upper() + ":")
                        __print_tabular(l)
            elif args.type == "macros":
                l = list_macros()
                if len(l) > 0:
                    __print_tabular(l)
            return 0
        # handle input file or stdin
        c =_input(args.infile)
        c = c.rstrip("\r\n") if isinstance(c, str) else c.rstrip(b"\r\n")
        # strip any other (CR)LF
        if args.strip:
            c = re.sub(r"\r?\n", "", c) if isinstance(c, str) else c.replace(b"\r\n", b"").replace(b"\n", b"")
        if args.command in ["decode", "encode"]:
            # encode or decode
            for encoding in args.encoding:
                c = getattr(codecs, ["encode", "decode"][args.command == "decode"])(c, encoding, args.errors)
            # handle output file or stdout
            if args.outfile:
                with open(args.outfile, 'wb') as f:
                    f.write(c)
            else:
                print(ensure_str(c or "Could not %scode :-(" % ["en", "de"][args.command == "decode"]), end="")
        elif args.command == "guess":
            r = codecs.guess(c,
                             getattr(stopfunc, args.stop_function, ["", "(?i)"][args.icase] + args.stop_function),
                             args.min_depth,
                             args.max_depth,
                             args.codec_categories,
                             args.exclude_codecs,
                             args.encoding,
                             not args.do_not_stop,
                             True,  # show
                             not args.no_heuristic,
                             args.extended,
                             args.verbose)
            for i, o in enumerate(r.items()):
                e, out = o
                if len(e) > 0:
                    if args.outfile:
                        n, ext = os.path.splitext(args.outfile)
                        fn = args.outfile if len(r) == 1 else "%s-%d%s" % (n, i+1, ext)
                    else:
                        print("Codecs: %s" % ", ".join(e))
                        print(ensure_str(out))
            if len(r) == 0:
                print("Could not decode :-(")
        elif args.command == "rank":
            for i, e in codecs.rank(c, args.extended, args.limit, args.codec_categories, args.exclude_codecs):
                s = "[+] %.5f: %s" % (i[0], e)
                print(s if len(s) <= 80 else s[:77] + "...")
    except Exception as e:
        m = str(e)
        print("codext: " + m[0].lower() + m[1:])




# -*- coding: UTF-8 -*-
"""Case Codecs - string hashing with SHAKE.

These are codecs for hashing strings, for use with other codecs in encoding chains.

These codecs:
- transform strings from str to str
- transform strings from bytes to bytes
- transform file content from str to bytes (write)
"""
import hashlib

from ..__common__ import add, b, PY3


if PY3:
    def shake_hash(i):
        def _hash_transform(l):
            l = (l or str(i)).lstrip("_-")
            def _encode(data, error="strict"):
                return getattr(hashlib, "shake_%d" % i)(b(data)).hexdigest(int(l)), len(data)
            return _encode
        return _hash_transform

    add("shake_128", shake_hash(128), pattern=r"^shake[-_]?128(|[-_][1-9]\d*)$", guess=None)
    add("shake_256", shake_hash(256), pattern=r"^shake[-_]?256(|[-_][1-9]\d*)$", guess=None)




# -*- coding: UTF-8 -*-
"""Case Codecs - string common checksums.

These are codecs for hashing strings, for use with other codecs in encoding chains.

These codecs:
- transform strings from str to str
- transform strings from bytes to bytes
- transform file content from str to bytes (write)
"""
from zlib import adler32

from ..__common__ import add, b


CRC = {
    '': {
        'a':             (0x1021, 0xc6c6, True, True, 0, 0xbf05),
    },
    8: {
        '':              (0x07, 0, False, False, 0, 0xf4),
        'aes':           (0x1d, 0xff, True, True, 0, 0x97),
        'autosar':       (0x2f, 0xff, False, False, 0xff, 0xdf),
        'bluetooth':     (0xa7, 0, True, True, 0, 0x26),
        'cdma2000':      (0x9b, 0xff, False, False, 0, 0xda),
        'dallas-1-wire': (0x31, 0, True, True, 0, 0xa1),
        'darc':          (0x39, 0, True, True, 0, 0x15),
        'dvb-s2':        (0xd5, 0, False, False, 0, 0xbc),
        'ebu':           (0x1d, 0xff, True, True, 0, 0x97),
        'gsm-a':         (0x1d, 0, False, False, 0, 0x37),
        'gsm-b':         (0x49, 0, False, False, 0xff, 0x94),
        'hitag':         (0x1d, 0xff, False, False, 0, 0xb4),
        'i-432-1':       (0x07, 0, False, False, 0x55, 0xa1),
        'i-code':        (0x1d, 0xfd, False, False, 0, 0x7e),
        'itu':           (0x07, 0, False, False, 0x55, 0xa1),
        'lte':           (0x9b, 0, False, False, 0, 0xea),
        'maxim':         (0x31, 0, True, True, 0, 0xa1),
        'maxim-dow':     (0x31, 0, True, True, 0, 0xa1),
        'mifare-mad':    (0x1d, 0xc7, False, False, 0, 0x99),
        'nrsc-5':        (0x31, 0xff, False, False, 0, 0xf7),
        'opensafety':    (0x2f, 0, False, False, 0, 0x3e),
        'rohc':          (0x07, 0xff, True, True, 0, 0xd0),
        'sae-j1850':     (0x1d, 0xff, False, False, 0xff, 0x4b),
        'smbus':         (0x07, 0, False, False, 0, 0xf4),
        'tech-3250':     (0x1d, 0xff, True, True, 0, 0x97),
        'wcdma':         (0x9b, 0, True, True, 0, 0x25),
    },
    10: {
        '':              (0x233, 0, False, False, 0, 0x199),
        'atm':           (0x233, 0, False, False, 0, 0x199),
        'cdma2000':      (0x3d9, 0x3ff, False, False, 0, 0x233),
        'gsm':           (0x175, 0, False, False, 0x3ff, 0x12a),
        'i-610':         (0x233, 0, False, False, 0, 0x199),
    },
    11: {
        '':              (0x385, 0x01a, False, False, 0, 0x5a3),
        'flexray':       (0x385, 0x01a, False, False, 0, 0x5a3),
        'umts':          (0x307, 0, False, False, 0, 0x061),
    },
    12: {
        '':              (0x80f, 0, False, True, 0, 0xdaf),
        '3gpp':          (0x80f, 0, False, True, 0, 0xdaf),
        'cdma2000':      (0xf13, 0xfff, False, False, 0, 0xd4d),
        'dect':          (0x80f, 0, False, False, 0, 0xf5b),
        'gsm':           (0xd31, 0, False, False, 0xfff, 0xb34),
        'umts':          (0x80f, 0, False, True, 0, 0xdaf),
    },
    13: {
        '':              (0x1cf5, 0, False, False, 0, 0x04fa),
        'bbc':           (0x1cf5, 0, False, False, 0, 0x04fa),
    },
    14: {
        '':              (0x0805, 0, True, True, 0, 0x082d),
        'darc':          (0x0805, 0, True, True, 0, 0x082d),
        'gsm':           (0x202d, 0, False, False, 0x3fff, 0x30ae),
    },
    15: {
        '':              (0x4599, 0, False, False, 0, 0x059e),
        'can':           (0x4599, 0, False, False, 0, 0x059e),
        'mpt1327':       (0x6815, 0, False, False, 1, 0x2566),
    },
    16: {
        'acorn':         (0x1021, 0, False, False, 0, 0x31c3),
        'arc':           (0x8005, 0, True, True, 0, 0xbb3d),
        'atom':          (0x002d, 0, True, True, 0, 0x4287),
        'aug-2-ccitt':   (0x1021, 0x84c0, False, False, 0, 0x19cf),
        'aug-2-citt':    (0x1021, 0x84c0, False, False, 0, 0x19cf),
        'aug-ccitt':     (0x1021, 0x1d0f, False, False, 0, 0xe5cc),
        'aug-citt':      (0x1021, 0x1d0f, False, False, 0, 0xe5cc),
        'autosar':       (0x1021, 0xffff, False, False, 0, 0x29b1),
        'bt-chip':       (0x1021, 0xffff, True, False, 0, 0x89f6),
        'buypass':       (0x8005, 0, False, False, 0, 0xfee8),
        'cms':           (0x8005, 0xffff, False, False, 0, 0xaee7),
        'ccitt':         (0x1021, 0, True, True, 0, 0x2189),
        'ccitt-false':   (0x1021, 0xffff, False, False, 0, 0x29b1),
        'ccitt-true':    (0x1021, 0, True, True, 0, 0x2189),
        'cdma2000':      (0xc867, 0xffff, False, False, 0, 0x4c06),
        'darc':          (0x1021, 0xffff, False, False, 0xffff, 0xd64e),
        'dds-110':       (0x8005, 0x800d, False, False, 0, 0x9ecf),
        'dect-r':        (0x0589, 0, False, False, 1, 0x007e),
        'dect-x':        (0x0589, 0, False, False, 0, 0x007f),
        'dnp':           (0x3d65, 0, True, True, 0xffff, 0xea82),
        'en-13757':      (0x3d65, 0, False, False, 0xffff, 0xc2b7),
        'epc':           (0x1021, 0xffff, False, False, 0xffff, 0xd64e),
        'epc-c1g2':      (0x1021, 0xffff, False, False, 0xffff, 0xd64e),
        'genibus':       (0x1021, 0xffff, False, False, 0xffff, 0xd64e),
        'gsm':           (0x1021, 0, False, False, 0xffff, 0xce3c),
        'i-code':        (0x1021, 0xffff, False, False, 0xffff, 0xd64e),
        'ibm':           (0x8005, 0, True, True, 0, 0xbb3d),
        'ibm-3740':      (0x1021, 0xffff, False, False, 0, 0x29b1),
        'ibm-sdlc':      (0x1021, 0xffff, True, True, 0xffff, 0x906e),
        'iec-61158-2':   (0x1dcf, 0xffff, False, False, 0xffff, 0xa819),
        'iso-hdlc':      (0x1021, 0xffff, True, True, 0xffff, 0x906e),
        'kermit':        (0x1021, 0, True, True, 0, 0x2189),
        'lha':           (0x8005, 0, True, True, 0, 0xbb3d),
        'lj1200':        (0x6f63, 0, False, False, 0, 0xbdf4),
        'maxim':         (0x8005, 0, True, True, 0xffff, 0x44c2),
        'maxim-dom':     (0x8005, 0, True, True, 0xffff, 0x44c2),
        'mcrf4xx':       (0x1021, 0xffff, True, True, 0, 0x6f91),
        'modbus':        (0x8005, 0xffff, True, True, 0, 0x4b37),
        'opensafety-a':  (0x5935, 0, False, False, 0, 0x5d38),
        'opensafety-b':  (0x755b, 0, False, False, 0, 0x20fe),
        'profibus':      (0x1dcf, 0xffff, False, False, 0xffff, 0xa819),
        'riello':        (0x1021, 0xb2aa, True, True, 0, 0x63d0),
        'spi-fujitsu':   (0x1021, 0x84c0, False, False, 0, 0x19cf),
        't10-dif':       (0x8bb7, 0, False, False, 0, 0xd0db),
        'teledisk':      (0xa097, 0, False, False, 0, 0x0fb3),
        'tms37157':      (0x1021, 0x89ec, True, True, 0, 0x26b1),
        'umts':          (0x8005, 0, False, False, 0, 0xfee8),
        'usb':           (0x8005, 0xffff, True, True, 0xffff, 0xb4c8),
        'v-41-lsb':      (0x1021, 0, True, True, 0, 0x2189),
        'verifone':      (0x8005, 0, False, False, 0, 0xfee8),
        'x-25':          (0x1021, 0xffff, True, True, 0xffff, 0x906e),
        'x-kermit':      (0x8408, 0, True, True, 0, 0x0c73),
        'x-xmodem':      (0x8408, 0, True, True, 0, 0x0c73),
        'xmodem':        (0x1021, 0, False, False, 0, 0x31c3),
        'zmodem':        (0x1021, 0, False, False, 0, 0x31c3),
    },
    17: {
        '':              (0x1685b, 0, False, False, 0, 0x04f03),
        'can-fd':        (0x1685b, 0, False, False, 0, 0x04f03),
    },
    21: {
        '':              (0x102899, 0, False, False, 0, 0x0ed841),
        'can-fd':        (0x102899, 0, False, False, 0, 0x0ed841),
    },
    24: {
        '':              (0x864cfb, 0xb704ce, False, False, 0, 0x21cf02),
        'ble':           (0x00065b, 0x555555, True, True, 0, 0xc25a56),
        'flexray-a':     (0x5d6dcb, 0xfedcba, False, False, 0, 0x7979bd),
        'flexray-b':     (0x5d6dcb, 0xabcdef, False, False, 0, 0x1f23b8),
        'interlaken':    (0x328b63, 0xffffff, False, False, 0xffffff, 0xb4f3e6),
        'lte-a':         (0x864cfb, 0, False, False, 0, 0xcde703),
        'lte-b':         (0x800063, 0, False, False, 0, 0x23ef52),
        'openpgp':       (0x864cfb, 0xb704ce, False, False, 0, 0x21cf02),
        'os-9':          (0x800063, 0xffffff, False, False, 0xffffff, 0x200fa5),
        'pgp':           (0x864cfb, 0xb704ce, False, False, 0, 0x21cf02),
    },
    30: {
        '':              (0x2030b9c7, 0x3fffffff, False, False, 0x3fffffff, 0x04c34abf),
        'cdma':          (0x2030b9c7, 0x3fffffff, False, False, 0x3fffffff, 0x04c34abf),
    },
    31: {
        '':              (0x04c11db7, 0x7fffffff, False, False, 0x7fffffff, 0x0ce9e46c),
        'philips':       (0x04c11db7, 0x7fffffff, False, False, 0x7fffffff, 0x0ce9e46c),
    },
    32: {
        '':              (0x04c11db7, 0xffffffff, True, True, 0xffffffff, 0xcbf43926),
        'aal5':          (0x04c11db7, 0xffffffff, False, False, 0xffffffff, 0xfc891918),
        'adccp':         (0x04C11db7, 0xffffffff, True, True, 0xffffffff, 0xcbf43926),
        'aixm':          (0x814141ab, 0, False, False, 0, 0x3010bf7f),
        'autosar':       (0xf4acfb13, 0xffffffff, True, True, 0xffffffff, 0x1697d06a),
        'b':             (0x04c11db7, 0xffffffff, False, False, 0xffffffff, 0xfc891918),
        'base91-c':      (0x1edc6f41, 0xffffffff, True, True, 0xffffffff, 0xe3069283),
        'base91-d':      (0xa833982b, 0xffffffff, True, True, 0xffffffff, 0x87315576),
        'bzip2':         (0x04c11db7, 0xffffffff, False, False, 0xffffffff, 0xfc891918),
        'c':             (0x1edc6f41, 0xffffffff, True, True, 0xffffffff, 0xe3069283),
        'castagnoli':    (0x1edc6f41, 0xffffffff, True, True, 0xffffffff, 0xe3069283),
        'cd-rom-edc':    (0x8001801b, 0, True, True, 0, 0x6ec2edc4),
        'cksum':         (0x04c11db7, 0, False, False, 0xffffffff, 0x765e7680),
        'd':             (0xa833982b, 0xffffffff, True, True, 0xffffffff, 0x87315576),
        'dect-b':        (0x04c11db7, 0xffffffff, False, False, 0xffffffff, 0xfc891918),
        'interlaken':    (0x1edc6f41, 0xffffffff, True, True, 0xffffffff, 0xe3069283),
        'iscsi':         (0x1edc6f41, 0xffffffff, True, True, 0xffffffff, 0xe3069283),
        'iso-hdlc':      (0x04c11db7, 0xffffffff, True, True, 0xffffffff, 0xcbf43926),
        'jamcrc':        (0x04c11db7, 0xffffffff, True, True, 0, 0x340bc6d9),
        'mpeg-2':        (0x04c11db7, 0xffffffff, False, False, 0, 0x0376e6e7),
        'mpeg2':         (0x04c11db7, 0xffffffff, False, False, 0, 0x0376e6e7),
        'posix':         (0x04c11db7, 0, False, False, 0xffffffff, 0x765e7680),
        'q':             (0x814141ab, 0, False, False, 0, 0x3010bf7f),
        'v-42':          (0x04C11db7, 0xffffffff, True, True, 0xffffffff, 0xcbf43926),
        'xfer':          (0x000000af, 0, False, False, 0, 0xbd0be338),
        'xz':            (0x04C11db7, 0xffffffff, True, True, 0xffffffff, 0xcbf43926),
        'zip':           (0x04C11DB7, 0xffffffff, True, True, 0xffffffff, 0xcbf43926),
    },
    40: {
        '':              (0x0004820009, 0, False, False, 0xffffffffff, 0xd4164fc646),
        'gsm':           (0x0004820009, 0, False, False, 0xffffffffff, 0xd4164fc646),
    },
    64: {
        '':              (0x000000000000001b, 0xffffffffffffffff, True, True, 0xffffffffffffffff, 0xb90956c775a41001),
        'ecma':          (0x42f0e1eba9ea3693, 0, False, False, 0, 0x6c40df5f0b497347),
        'ecma-182':      (0x42f0e1eba9ea3693, 0, False, False, 0, 0x6c40df5f0b497347),
        'go-ecma':       (0x42f0e1eba9ea3693, 0xffffffffffffffff, True, True, 0xffffffffffffffff, 0x995dc9bbdf1939fa),
        'go-iso':        (0x000000000000001b, 0xffffffffffffffff, True, True, 0xffffffffffffffff, 0xb90956c775a41001),
        'iso':           (0x000000000000001b, 0xffffffffffffffff, True, True, 0xffffffffffffffff, 0xb90956c775a41001),
        'we':            (0x42f0e1eba9ea3693, 0xffffffffffffffff, False, False, 0xffffffffffffffff, 0x62ec59e3f1a4f00a),
        'xz':            (0x42f0e1eba9ea3693, 0xffffffffffffffff, True, True, 0xffffffffffffffff, 0x995dc9bbdf1939fa),
        'xz64':          (0x42f0e1eba9ea3693, 0xffffffffffffffff, True, True, 0xffffffffffffffff, 0x995dc9bbdf1939fa),
    },
    82: {
        '':              (0x0308c0111011401440411, 0, True, True, 0, 0x09ea83f625023801fd612),
        'darc':          (0x0308c0111011401440411, 0, True, True, 0, 0x09ea83f625023801fd612),
    },
}

_pattern = lambda n="": r"^crc" + str(n) + r"(|[-_]?(?:%s))$" % "|".join(x for x in CRC[n].keys() if len(x) > 0)
_rev_int = lambda i, l=None: int(bin(i)[2:].zfill(l or len(bin(i)[2:]))[::-1], 2)


def crc(data, length, poly, init=0, refl_in=False, refl_out=False, xor_out=0):
    """ Generic CRC computation function. """
    table = [None] * 256
    # input reflected
    if refl_in:
        init, poly = _rev_int(init, length), _rev_int(poly, length)
        # prepare the lookup table
        for x in range(2**8):
            crc = x
            for i in range(8):
                crc = (crc >> 1) ^ [0, poly][crc & 0x1 != 0]
            table[x] = crc
        # compute CRC
        crc = init
        for c in data:
            crc = (crc >> 8) ^ table[(crc ^ ord(c)) & 0xff]
    # input NOT reflected
    else:
        # prepare the lookup table
        for x in range(2**8):
            crc = x << (length - 8)
            for i in range(8):
                crc = ((crc << 1) & ((1 << length) - 1)) ^ [0, poly][crc >> (length - 1) != 0]
            table[x] = crc
        # compute CRC
        crc = init
        for c in data:
            crc = ((crc << 8) & ((1 << length) - 1)) ^ table[(crc >> (length - 8)) ^ ord(c)]
    # output reflected
    if refl_in ^ refl_out:
        crc = _rev_int(crc, length)
    return crc ^ xor_out


def crc_checksum(n=""):
    def _crc(name):
        def _encode(data, error="strict"):
            r = crc(data, n or 16, *CRC[n][name.lstrip("-_")][:5])
            return "%0{}x".format(round((n or 16)/4+.5)) % r, len(data)
        return _encode
    return _crc


add("adler32", lambda data, error="strict": (adler32(b(data)) & 0xffffffff, len(data)), guess=None)
add("crca", crc_checksum(), pattern=_pattern(), guess=None)
for i in CRC.keys():
    if isinstance(i, int):
        add("crc%d" % i, crc_checksum(i), pattern=_pattern(i), guess=None)

